@@ -21,9 +21,17 @@ use Errno qw( EAGAIN ETIMEDOUT );
 use Carp;
 use strict;
 
-# from openssl/ssl.h, should be better in Net::SSLeay
-use constant SSL_SENT_SHUTDOWN => 1;
-use constant SSL_RECEIVED_SHUTDOWN => 2;
+use constant {
+	SSL_VERIFY_NONE => Net::SSLeay::VERIFY_NONE(),
+	SSL_VERIFY_PEER => Net::SSLeay::VERIFY_PEER(),
+	SSL_VERIFY_FAIL_IF_NO_PEER_CERT => Net::SSLeay::VERIFY_FAIL_IF_NO_PEER_CERT(),
+	SSL_VERIFY_CLIENT_ONCE => Net::SSLeay::VERIFY_CLIENT_ONCE(),
+	# from openssl/ssl.h, should be better in Net::SSLeay
+	SSL_SENT_SHUTDOWN => 1,
+	SSL_RECEIVED_SHUTDOWN => 2,
+};
+
+	
 
 # non-XS Versions of Scalar::Util will fail
 BEGIN{
@@ -45,7 +53,11 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 	my $y = Net::SSLeay::ERROR_WANT_WRITE();
 	use constant SSL_WANT_WRITE => dualvar( \$y, 'SSL wants a write first' );
 
-	@EXPORT = qw( SSL_WANT_READ SSL_WANT_WRITE $SSL_ERROR GEN_DNS GEN_IPADD );
+	@EXPORT = qw( 
+		SSL_WANT_READ SSL_WANT_WRITE SSL_VERIFY_NONE SSL_VERIFY_PEER 
+		SSL_VERIFY_FAIL_IF_NO_PEER_CERT SSL_VERIFY_CLIENT_ONCE
+		$SSL_ERROR GEN_DNS GEN_IPADD 
+	);
 }
 
 my @caller_force_inet4; # in case inet4 gets forced we store here who forced it
@@ -66,7 +78,7 @@ BEGIN {
 	}) {
 		@ISA = qw(IO::Socket::INET);
 	}
-	$VERSION = '1.22';
+	$VERSION = '1.31';
 	$GLOBAL_CONTEXT_ARGS = {};
 
 	#Make $DEBUG another name for $Net::SSLeay::trace
@@ -206,20 +218,28 @@ sub configure_SSL {
 	my ($self, $arg_hash) = @_;
 
 	my $is_server = $arg_hash->{'SSL_server'} || $arg_hash->{'Listen'} || 0;
+
 	my %default_args = (
 		Proto => 'tcp',
 		SSL_server => $is_server,
-		SSL_ca_file => 'certs/my-ca.pem',
-		SSL_ca_path => 'ca/',
 		SSL_use_cert => $is_server,
 		SSL_check_crl => 0,
 		SSL_version	=> 'sslv23',
-		SSL_verify_mode => Net::SSLeay::VERIFY_NONE(),
+		SSL_verify_mode => SSL_VERIFY_NONE,
 		SSL_verify_callback => undef,
 		SSL_verifycn_scheme => undef,  # don't verify cn
 		SSL_verifycn_name => undef,    # use from PeerAddr/PeerHost
 	);
-	 
+
+	# common problem forgetting SSL_use_cert
+	# if client cert is given but SSL_use_cert undef assume that it 
+	# should be set
+	if ( ! $is_server && ! defined $arg_hash->{SSL_use_cert} 
+		&& ( grep { $arg_hash->{$_} } qw(SSL_cert SSL_cert_file)) 
+		&& ( grep { $arg_hash->{$_} } qw(SSL_key SSL_key_file)) ) {
+		$arg_hash->{SSL_use_cert} = 1 
+	}
+
 	# SSL_key_file and SSL_cert_file will only be set in defaults if 
 	# SSL_key|SSL_key_file resp SSL_cert|SSL_cert_file are not set in
 	# $args_hash
@@ -231,19 +251,21 @@ sub configure_SSL {
 			:  "certs/client-${k}.pem";
 	}	
 
+	# add only SSL_ca_* if not in args
+	if ( ! exists $arg_hash->{SSL_ca_file} && ! exists $arg_hash->{SSL_ca_path} ) {
+		if ( -f 'certs/my-ca.pem' ) {
+			$default_args{SSL_ca_file} = 'certs/my-ca.pem'
+		} elsif ( -d 'ca/' ) {
+			$default_args{SSL_ca_path} = 'ca/'
+		}
+	}
+		
 	#Replace nonexistent entries with defaults
 	%$arg_hash = ( %default_args, %$GLOBAL_CONTEXT_ARGS, %$arg_hash );
 
 	#Avoid passing undef arguments to Net::SSLeay
 	defined($arg_hash->{$_}) or delete($arg_hash->{$_}) foreach (keys %$arg_hash);
 
-	#Handle CA paths properly if no CA file is specified
-	if ($arg_hash->{'SSL_ca_path'} ne '' and !(-f $arg_hash->{'SSL_ca_file'})) {
-		DEBUG(1, "CA file $arg_hash->{SSL_ca_file} not found, using CA path instead.\n" )
-			if $arg_hash->{SSL_ca_file};
-		$arg_hash->{'SSL_ca_file'} = '';
-	}
-
 	my $vcn_scheme = delete $arg_hash->{SSL_verifycn_scheme};
 	if ( $vcn_scheme && $vcn_scheme ne 'none' ) {
 		# don't access ${*self} inside callback - this seems to create
@@ -253,7 +275,7 @@ sub configure_SSL {
 		my $host = $arg_hash->{SSL_verifycn_name};
 		if (not defined($host)) {
 			if ( $host = $arg_hash->{PeerAddr} || $arg_hash->{PeerHost} ) {
-				$host =~s{:\w+$}{};
+				$host =~s{:[a-zA-Z0-9_\-]+$}{};
 			}
 		}
 		$host ||= ref($vcn_scheme) && $vcn_scheme->{callback} && 'unknown';
@@ -645,11 +667,14 @@ sub generic_write {
 	my $written;
 	if ( $write_all ) {
 		my $data = $length < $buf_len-$offset ? substr($$buffer, $offset, $length) : $$buffer;
-		$written = Net::SSLeay::ssl_write_all($ssl, $data);
+		($written, my $errs) = Net::SSLeay::ssl_write_all($ssl, $data);
+		# ssl_write_all returns number of bytes written
+		$written = undef if ! $written && $errs;
 	} else {
 		$written = Net::SSLeay::write_partial( $ssl,$offset,$length,$$buffer );
+		# write_partial does SSL_write which returns -1 on error
+		$written = undef if $written < 0;
 	}
-	$written = undef if $written < 0; # Net::SSLeay::write returns -1 not undef on error
 	if ( !defined($written) ) {
 		$self->_set_rw_error( $ssl,-1 )
 			|| $self->error("SSL write error");
@@ -784,8 +809,8 @@ sub close {
 sub stop_SSL {
 	my $self = shift || return _invalid_object();
 	my $stop_args = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
-	return $self->error("SSL object already closed") 
-		unless (${*$self}{'_SSL_opened'} == 1);
+	return $self->error("SSL object not open") 
+		if ! ${*$self}{'_SSL_opened'};
 
 	if (my $ssl = ${*$self}{'_SSL_object'}) {
 		my $shutdown_done;
@@ -859,7 +884,7 @@ sub stop_SSL {
 sub kill_socket {
 	my $self = shift;
 	shutdown($self, 2);
-	$self->close(SSL_no_shutdown => 1) if (${*$self}{'_SSL_opened'} == 1);
+	$self->close(SSL_no_shutdown => 1) if ${*$self}{'_SSL_opened'};
 	delete(${*$self}{'_SSL_ctx'});
 	return;
 }
@@ -1037,6 +1062,7 @@ sub dump_peer_certificate {
 	$scheme{imap} = $scheme{ldap}; # rfc 2595
 	$scheme{acap} = $scheme{ldap}; # rfc 2595
 	$scheme{nntp} = $scheme{ldap}; # rfc 4642
+	$scheme{ftp}  = $scheme{http}; # rfc 4217
 
 	# function to verify the hostname
 	#
@@ -1076,8 +1102,9 @@ sub dump_peer_certificate {
 			 # definitly no hostname, try IPv4
 			$ip4 = inet_aton( $identity ) or croak "'$identity' is not IPv4, but neither IPv6 nor hostname";
 		} else {
-			# assume hostname
-			if ( $identity !~m{^[a-zA-Z0-9-_\.]+$} ) {
+			# assume hostname, check for umlauts etc
+			if ( $identity =~m{[^a-zA-Z0-9_.\-]} ) {
+				$identity =~m{\0} and croak("name '$identity' has \\0 byte");
 				$identity = idn_to_ascii($identity) or
 					croak "Warning: Given name '$identity' could not be converted to IDNA!";
 			}
@@ -1094,12 +1121,12 @@ sub dump_peer_certificate {
 			# The RFCs are in this regard unspecific but we don't want to have to
 			# deal with certificates like *.com, *.co.uk or even *
 			# see also http://nils.toedtmann.net/pub/subjectAltName.txt
-			if ( $wtyp eq 'anywhere' and $name =~m{^([\w\-]*)\*(.+)} ) {
-				$pattern = qr{^\Q$1\E[\w\-]*\Q$2\E$}i;
+			if ( $wtyp eq 'anywhere' and $name =~m{^([a-zA-Z0-9_\-]*)\*(.+)} ) {
+				$pattern = qr{^\Q$1\E[a-zA-Z0-9_\-]*\Q$2\E$}i;
 			} elsif ( $wtyp eq 'leftmost' and $name =~m{^\*(\..+)$} ) {
-				$pattern = qr{^[\w\-]*\Q$1\E$}i;
+				$pattern = qr{^[a-zA-Z0-9_\-]*\Q$1\E$}i;
 			} else {
-				$pattern = qr{^\Q$name}i;
+				$pattern = qr{^\Q$name\E$}i;
 			}
 			return $identity =~ $pattern;
 		};
@@ -1183,7 +1210,7 @@ sub error {
 sub DESTROY {
 	my $self = shift || return;
 	$self->close(_SSL_in_DESTROY => 1, SSL_no_shutdown => 1) 
-		if (${*$self}{'_SSL_opened'} == 1);
+		if ${*$self}{'_SSL_opened'};
 	delete(${*$self}{'_SSL_ctx'});
 }
 
@@ -1292,6 +1319,7 @@ sub CLOSE {							 #<---- Do not change this function!
 
 
 package IO::Socket::SSL::SSL_Context;
+use Carp;
 use strict;
 
 my %CTX_CREATED_IN_THIS_THREAD;
@@ -1341,18 +1369,32 @@ sub new {
 
 
 	my $verify_mode = $arg_hash->{SSL_verify_mode};
-	unless ($verify_mode == Net::SSLeay::VERIFY_NONE()) {
-		Net::SSLeay::CTX_load_verify_locations(
-			$ctx, $arg_hash->{SSL_ca_file},$arg_hash->{SSL_ca_path}
-		) || return IO::Socket::SSL->error("Invalid certificate authority locations");
+	if ( $verify_mode != Net::SSLeay::VERIFY_NONE() and 
+		! Net::SSLeay::CTX_load_verify_locations( 
+			$ctx, $arg_hash->{SSL_ca_file} || '',$arg_hash->{SSL_ca_path} || '') ) {
+		if ( ! $arg_hash->{SSL_ca_file} && ! $arg_hash->{SSL_ca_path} ) {
+			carp("No certificate verification because neither SSL_ca_file nor SSL_ca_path known");
+			$verify_mode = Net::SSLeay::VERIFY_NONE();
+		} else {
+			return IO::Socket::SSL->error("Invalid certificate authority locations");
+		}
 	}
 
 	if ($arg_hash->{'SSL_check_crl'}) {
 		if (Net::SSLeay::OPENSSL_VERSION_NUMBER() >= 0x0090702f) {
-			Net::SSLeay::X509_STORE_set_flags(
-				Net::SSLeay::CTX_get_cert_store($ctx),
-				Net::SSLeay::X509_V_FLAG_CRL_CHECK()
-			);
+                    Net::SSLeay::X509_STORE_set_flags(
+                        Net::SSLeay::CTX_get_cert_store($ctx),
+                        Net::SSLeay::X509_V_FLAG_CRL_CHECK()
+                      );
+                    if ($arg_hash->{'SSL_crl_file'}) {
+                        my $bio = Net::SSLeay::BIO_new_file($arg_hash->{'SSL_crl_file'}, 'r');
+                        my $crl = Net::SSLeay::PEM_read_bio_X509_CRL($bio);
+                        if ( $crl ) {
+                            Net::SSLeay::X509_STORE_add_crl(Net::SSLeay::CTX_get_cert_store($ctx), $crl);
+                        } else {
+                            return IO::Socket::SSL->error("Invalid certificate revocation list");
+                        }
+                    }
 		} else {
 			return IO::Socket::SSL->error("CRL not supported for OpenSSL < v0.9.7b");
 		}
@@ -1446,6 +1488,7 @@ sub session_cache {
 	my $ctx = shift;
 	my $cache = $ctx->{'session_cache'} || return;
 	my ($addr,$port,$session) = @_;
+	$port ||= $addr =~s{:(\w+)$}{} && $1; # host:port
 	my $key = "$addr:$port";
 	return defined($session) 
 		? $cache->add_session($key, $session)
@@ -1613,6 +1656,16 @@ If this is set, it forces IO::Socket::SSL to use a certificate and key, even if
 you are setting up an SSL client.  If this is set to 0 (the default), then you will
 only need a certificate and key if you are setting up a server.
 
+SSL_use_cert will implicitly be set if SSL_server is set.
+For convinience it is also set if it was not given but a cert was given for use
+(SSL_cert_file or similar).
+
+=item SSL_server
+
+Use this, if the socket should be used as a server.
+If this is not explicitly set it is assumed, if Listen with given when creating
+the socket.
+
 =item SSL_key_file
 
 If your RSA private key is not in default place (F<certs/server-key.pem> for servers,
@@ -1711,11 +1764,18 @@ with B<start_SSL> the name has to be given in B<SSL_verifycn_name>.
 
 =item SSL_check_crl
 
-If you want to verify that the peer certificate has not been revoked by the
-signing authority, set this value to true.	OpenSSL will search for the CRL
-in your SSL_ca_path.  See the Net::SSLeay documentation for more details.
-Note that this functionality appears to be broken with OpenSSL < v0.9.7b,
-so its use with lower versions will result in an error.
+If you want to verify that the peer certificate has not been revoked
+by the signing authority, set this value to true. OpenSSL will search
+for the CRL in your SSL_ca_path, or use the file specified by
+SSL_crl_file.  See the Net::SSLeay documentation for more details.
+Note that this functionality appears to be broken with OpenSSL <
+v0.9.7b, so its use with lower versions will result in an error.
+
+=item SSL_crl_file
+
+If you want to specify the CRL file to be used, set this value to the
+pathname to be used.  This must be used in addition to setting
+SSL_check_crl.
 
 =item SSL_reuse_ctx
 
@@ -1745,7 +1805,7 @@ added.
 
 Specifies session cache object which should be used instead of creating a new.
 Overrules SSL_session_cache_size.
-This option is useful if you wan't to reuse the cache, but not the rest of
+This option is useful if you want to reuse the cache, but not the rest of
 the context.
 
 A session cache object can be created using 
